%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This function performs ordinary cross validation (leaving-one-out) to
% compute the optimal value of the regularization parameter.
%
% Written and compiled by Biswadip Dey, University of Maryland in May 2012.
%
%
% NOTE: As the cross validation process for computing optimal value for the
% regularization parameter is parallel in nature, we extract benefit from
% multi-core processor and Matlab Parallel Computing Toolbox (PCT). If you
% are equipped with a multi-core computer and Matlab Parallel Computing
% Toolbox, then by modifying (commenting and uncommenting) a couple of 
% lines in this function you can make the program much faster. The factor 
% of reduction in execution time is approximately same as the number of 
% available cores in your computer. The details about this modification is 
% provided within the code (Lines: 51-58, 69 and 72).
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% **************************** Initialization ****************************
%
function [Lambda_Array,Recon_Error,OCV_Cost_Array] = Ord_Crss_Validation...
    (Curve,Time_Array,FrameRate,Lambda_Array)
%
%****** Curve >>>>>> Data is Arranged HORIZONTALLY. (in a 3xN matrix, N is
%                    the number of data points)
%****** Time_Array >>>>>> Time Tags (They are basically serial numbers for 
%                         the snapshots, and is used to keep a track of
%                         missing frames, if any.)
%****** FrameRate >>>>>> Image capture rate (fps)
%****** Lambda_Array >>>>>> Stores the values for the regularizing
%                           parameter for which we compute the cross 
%                           validation cost.
%%
tic;
%
%
Dim_1_Recon_Error = length(Lambda_Array);
Dim_2_Recon_Error = length(Time_Array);
%
Recon_Error = zeros(Dim_1_Recon_Error,Dim_2_Recon_Error);
%
Absolute_TIME_Global = (Time_Array - Time_Array(1))*(1/FrameRate);
%
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% ************************************************************************
%
% UNCOMMENT NEXT 8 (51-58) LINES IF YOU HAVE MULTI-CORE COMPUTER AND MATLAB
% PARALLEL COMPUTING TOOLBOX (PCT), OTHERWISE KEEP IT COMMENTED.
% NOTE: Replace "12" by number of available cores in the multicore computer
%
% if matlabpool('size')==0
%     matlabpool open 12
% else
%     if matlabpool('size')<12
%         matlabpool close
%         matlabpool open 12
%     end
% end
%
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% ************************************************************************
%
for lambda_idx = 1:Dim_1_Recon_Error
    %
    Lambda = Lambda_Array(lambda_idx);
    %
    % UNCOMMENT THE NEXT (69) LINE IF YOU HAVE A SINGLE-CORE COMPUTER.
    for LeftOut_idx = 2:Dim_2_Recon_Error-1
    %
    % UNCOMMENT THE NEXT (72) LINE IF YOU HAVE A MULTI-CORE COMPUTER.
%     parfor LeftOut_idx = 2:Dim_2_Recon_Error-1
        %% ****************************************************************
        %
        % Removes the point to be left-out from the set used for
        % prediction.
        Temp_Curve = [Curve(:,1:LeftOut_idx-1) Curve(:,LeftOut_idx+1:length(Time_Array))];
        Temp_Time = [Time_Array(1:LeftOut_idx-1) Time_Array(LeftOut_idx+1:length(Time_Array))];
        %
        % Reconstructs the trajectory by calling the function 
        % "Linear_Filter.m" 
        [~,POS,~,~,~,~,~,Time_HR] = Linear_Filter(Lambda,1,Temp_Curve,Temp_Time,FrameRate);
        %
        idx_POS = find(abs(Absolute_TIME_Global(LeftOut_idx)-Time_HR)<1e-14);
        %
        % Computes the prediction error at the left-out point.
        MisMatch_in_Fit = POS(:,idx_POS) - Curve(:,LeftOut_idx);
        Recon_Error(lambda_idx,LeftOut_idx) = MisMatch_in_Fit'*MisMatch_in_Fit;
        %
        % Displays a message showing the ongoing cross validation process
        if rem(LeftOut_idx,100)==0
            display('Cross Validation is running .....');            
        end
    end
    %
    fprintf('\n');
    display(strcat('Completed Lambda - (',num2str(lambda_idx),')/(',num2str(length(Lambda_Array)),')'));
    toc;
    fprintf('\n\n');
    %
end
%
%
OCV_Cost_Array = sum(Recon_Error,2)/(length(Time_Array)-2);